from dataclasses import dataclass, field


@dataclass
class HourObject:
    nh: list = field(default_factory=lambda: [])
    ch: list = field(default_factory=lambda: [])
    dyn_ch: dict = field(default_factory=lambda: {})
    offset_dict: dict = field(default_factory=lambda: {})
    pricedict: dict = field(default_factory=lambda: {})

    def remove_cheap_hours(self, min_price:float = 0):
        lst = (h for h in self.pricedict if self.pricedict[h] < min_price)
        for h in lst:
            if h in self.nh:
                self.nh.remove(h)
            elif h in self.ch:
                self.ch.remove(h)
                self.dyn_ch.pop(h)    
        return self

    def _add(self, lst: list):
        for h in lst:
            if h not in self.nh:
                self.nh.append(h)
                if h in self.ch:
                    self.ch.remove(h)
                if len(self.dyn_ch):
                    if h in self.dyn_ch.keys():
                        self.dyn_ch.pop(h)

    def add_expensive_hours(self, max_price: float = 0):
        lst = (h for h in self.pricedict if self.pricedict[h] >= max_price)
        self._add(lst)
        self.nh.sort()
        return self



from ....services.hourselection.hourselectionservice.hoursselection_helpers import try_remove
from .hourobject import HourObject

def update_interim_lists(_range: range, old: HourObject, new, index_devidation: int):
        _new = _convert_collections(new, index_devidation)
        try:
            for i in _range:
                if i in _new.nh:
                    if i not in old.nh:
                        old.nh.append(i)
                        old.ch = try_remove(i, old.ch)
                        old.dyn_ch = try_remove(i, old.dyn_ch)
                elif i in _new.ch:
                    if i not in old.ch:
                        old.ch.append(i)
                        old.dyn_ch[i] = _new.dyn_ch[i]
                        old.nh = try_remove(i, old.nh)
                else:
                    old.nh = try_remove(i, old.nh)
                    old.ch = try_remove(i, old.ch)
                    old.dyn_ch = try_remove(i, old.dyn_ch)
                old.nh.sort()
                old.ch.sort()
                old.dyn_ch = dict(sorted(old.dyn_ch.items()))
        except IndexError:
            raise IndexError("Error on updating interim lists.")
        
        for r in _new.offset_dict.keys():
            old.offset_dict[r] = _new.offset_dict[r]
            old.pricedict[r] = _new.pricedict[r]
        return old
    
    
def _convert_collections(new: HourObject, index_deviation: int) -> HourObject:
    """Converts the hourobject-collections to interim days, based on the index-deviation provided."""
    total_length = len(new.pricedict.keys())
    ret = HourObject([], [], {})
    ret.nh = _chop_list(new.nh, index_deviation,total_length)
    ret.ch = _chop_list(new.ch, index_deviation, total_length)
    ret.dyn_ch = _chop_dict(new.dyn_ch, index_deviation, total_length)
    ret.offset_dict = _chop_dict(new.offset_dict, index_deviation, total_length)
    ret.pricedict = _chop_dict(new.pricedict, index_deviation, total_length)
    return ret


def _chop_list(lst: list, index_deviation:int, total_length:int = 24):
        return [n+index_deviation for n in lst if 0 <= n+index_deviation < total_length]

def _chop_dict(dct: dict, index_deviation: int, total_length: int = 24):
    return {key+index_deviation:value for (key,value) in dct.items() if 0 <= key+index_deviation < total_length}



from dataclasses import dataclass, field
from typing import List
from .hours_model import HoursModel
from .hourselection_options import HourSelectionOptions
import logging

_LOGGER = logging.getLogger(__name__)
        

@dataclass(frozen=False)
class HourSelectionModel:
    prices_today: List[float] = field(default_factory=lambda : [])
    prices_tomorrow: List[float] = field(default_factory=lambda : [])
    adjusted_average: float = None
    current_peak: float = 0.0
    hours: HoursModel = HoursModel()
    options: HourSelectionOptions = HourSelectionOptions

    def __post_init__(self):
        self.validate()

    def validate(self):
        assert 0 < self.options.cautionhour_type <= 1
        assert isinstance(self.prices_today, list)
        assert isinstance(self.prices_tomorrow, list)

        if isinstance(self.adjusted_average, (int, float)):
            assert self.adjusted_average >= 0
        else:
            assert self.adjusted_average is None
        


from datetime import datetime
from ...util import _convert_quarterly_minutes
from ...PeaqErrors import PeaqValueError


class Prediction:
    def __init__(self, hub = None):
        self._hub = hub

    @property
    def predictedenergy(self) -> float:
        return Prediction._predicted_energy(
            datetime.now().minute,
            datetime.now().second,
            self._hub.sensors.powersensormovingaverage.value if hasattr(self._hub.sensors, "powersensormovingaverage") else 0,
            self._hub.sensors.totalhourlyenergy.value,
            self._hub.sensors.locale.data.is_quarterly(self._hub.sensors.locale.data)
        )

    @property
    def predictedpercentageofpeak(self) -> float:
        if hasattr(self._hub.sensors, "current_peak"):
            return Prediction._predicted_percentage_of_peak(
                self._hub.sensors.current_peak.value,
                self.predictedenergy
            )
        return 0.0

    @staticmethod
    def _predicted_energy(
            now_min: int,
            now_sec: int,
            power_avg: float,
            total_hourly_energy: float,
            is_quarterly:bool=False
    ) -> float:
        """This method does allow for negative energy, ie from solar panels being read."""
        if now_min not in range(0, 60):
            raise PeaqValueError(f"Value 'now_min' ({now_min}) must be between (0..60]")
        if now_sec not in range(0, 60):
            raise PeaqValueError(f"Value 'now_max' ({now_sec}) must be between (0..60]")
        
        minute = _convert_quarterly_minutes(now_min, is_quarterly)

        if total_hourly_energy != 0 and (minute > 0 or (minute + now_sec) > 30):
            ret = (((power_avg / 60 / 60) * (3600 - ((minute * 60) + now_sec)) + total_hourly_energy * 1000) / 1000)
        else:
            ret = power_avg / 1000
        return round(ret, 3)

    @staticmethod
    def _predicted_percentage_of_peak(
            peak: float,
            predicted_energy: float
    ) -> float:
        if peak == 0.0 or peak is None:
            return 0
        elif predicted_energy <= 0.0 or predicted_energy is None:
            return 0
        return round((predicted_energy / peak) * 100, 2)



from dataclasses import dataclass, field
from datetime import datetime
from abc import abstractmethod
from ..const import NON_HOUR, CAUTION_HOUR, CHARGING_PERMITTED
from ...timer.timer import Timer
from ...scheduler.scheduler import Scheduler

class Hours:
    _timer: Timer
    _scheduler: Scheduler

    def __init__(
            self,
            price_aware: bool,
            non_hours: list = None,
            caution_hours: list = None
    ):
        self._non_hours = non_hours
        self._caution_hours = caution_hours
        self._price_aware = price_aware
        self._is_initialized = False

    @property
    def state(self) -> str:
        if datetime.now().hour in self.non_hours:
            return NON_HOUR
        if datetime.now().hour in self.caution_hours:
            return CAUTION_HOUR
        return CHARGING_PERMITTED

    @property
    def price_aware(self) -> bool:
        return self._price_aware

    @property
    def timer(self) -> Timer:
        return self._timer

    @property
    def scheduler(self) -> Scheduler:
        return self._scheduler

    @property
    @abstractmethod
    def non_hours(self):
        pass

    @property
    @abstractmethod
    def is_initialized(self):
        pass

    @non_hours.setter
    @abstractmethod
    def non_hours(self, val):
        pass

    @property
    @abstractmethod
    def caution_hours(self):
        pass

    @caution_hours.setter
    @abstractmethod
    def caution_hours(self, val):
        pass

    @property
    @abstractmethod
    def nordpool_entity(self):
        pass

    @abstractmethod
    def update_nordpool(self) -> None:
        pass

    @abstractmethod
    async def async_update_top_price(self, dyn_top_price) -> None: 
        pass

    @property
    @abstractmethod
    def dynamic_caution_hours(self) -> dict:
        pass

    @abstractmethod
    def update_prices(self, prices:dict = [], prices_tomorrow:dict=[]) -> None:
        pass

    @property
    @abstractmethod
    def options(self):
        pass

    @property
    @abstractmethod
    def offsets(self) -> dict:
        pass

    @property
    @abstractmethod
    def prices(self) -> list:
        pass

    @prices.setter
    @abstractmethod
    def prices(self, val):
        pass

    @property
    @abstractmethod
    def prices_tomorrow(self) -> list:
        pass

    @prices_tomorrow.setter
    @abstractmethod
    def prices_tomorrow(self, val):
        pass

    @property
    @abstractmethod
    def absolute_top_price(self):
        pass

    @property
    @abstractmethod
    def min_price(self):
        pass

    @property
    @abstractmethod
    def cautionhour_type_string(self) -> str:
        pass

    @property
    @abstractmethod
    def adjusted_average(self) -> float:
        pass

    @adjusted_average.setter
    @abstractmethod
    def adjusted_average(self, val):
        pass

    @abstractmethod
    async def async_get_average_kwh_price(self):
        pass

    @abstractmethod
    async def async_get_total_charge(self):
        pass
    


import statistics as stat
import logging
from ....models.hourselection.hourobjects.hourobject import HourObject


_LOGGER = logging.getLogger(__name__)


def create_dict(input: list):
    ret = {}
    for idx, val in enumerate(input):
        ret[idx] = val
    if 23 <= len(ret) <= 24:
        return ret
    elif len(ret) == 25:
        _LOGGER.debug(f"Looks like we are heading into DST. re-parsing hours")
        input.pop(2)
        return create_dict(input)
    else:
        _LOGGER.exception(f"Could not create dictionary from pricelist: {input} with len {len(ret)}.")
        raise ValueError


def convert_none_list(lst: any) -> list:
    ret = []
    if lst is None or not isinstance(lst, list):
        return ret
    try:
        for l in lst:
            if l is None:
                return ret
        return lst
    except:
        return _make_array_from_empty(lst)


def _try_parse(input:str, parsetype:type):
    try:
        ret = parsetype(input)
        return ret
    except:
        return False


def _make_array_from_empty(input: str) -> list:
    array = input.split(",")
    list = [p for p in array if len(p)]
    ret = []
    if len(list) > 24:
        try:
            for l in list:
                parsed_item = _try_parse(l, float)
                if not parsed_item:
                    parsed_item = _try_parse(l, int)
                assert isinstance(parsed_item, (float,int))
                ret.append(parsed_item)
            return ret
        except:
            _LOGGER.warning("Unable to create empty list for prices.")
            pass
    return []
    
def try_remove(value, collection: list|dict):
    if isinstance(collection, dict):
        if value in collection.keys():
            collection.pop(value)
    elif isinstance(collection, list):
        if value in collection:
            collection.remove(value)
    return collection


import logging
from ...models.hourselection.cautionhourtype import CautionHourType
from ...models.hourselection.hourselection_model import HourSelectionModel
from ...models.hourselection.hourselection_options import HourSelectionOptions
from ...models.hourselection.hourtypelist import HourTypeList
from .hourselectionservice.hourselectionservice import HourSelectionService
from .hourselectionservice.hoursselection_helpers import convert_none_list
_LOGGER = logging.getLogger(__name__)


class Hoursselection:
    def __init__(
            self,      
            absolute_top_price: float = 0,
            min_price: float = 0,
            cautionhour_type: str = CautionHourType.SUAVE.value,
            blocknocturnal: bool = False,
            base_mock_hour: int = None
    ):
        self.cautionhour_type_enum = CautionHourType(cautionhour_type.lower()) if isinstance(cautionhour_type, str) else cautionhour_type
        self.model = HourSelectionModel(
            options=HourSelectionOptions(
                cautionhour_type=CautionHourType.get_num_value(cautionhour_type), 
                min_price=min_price,
                top_price= absolute_top_price,
                blocknocturnal= blocknocturnal
                    )
                )
        self.model.validate()
        self.service = HourSelectionService(parent=self, base_mock_hour=base_mock_hour)
    
    @property
    def offsets(self) -> dict:
        return self.model.hours.offset_dict

    @property
    def non_hours(self) -> list:
        self.model.hours.update_hour_lists(listtype=HourTypeList.NonHour, hour=self.service.set_hour())
        return self.model.hours.non_hours

    @property
    def caution_hours(self) -> list:
        self.model.hours.update_hour_lists(listtype=HourTypeList.CautionHour, hour=self.service.set_hour())
        return self.model.hours.caution_hours

    @property
    def dynamic_caution_hours(self) -> dict:
        self.model.hours.update_hour_lists(listtype=HourTypeList.DynCautionHour, hour=self.service.set_hour())
        return self.model.hours.dynamic_caution_hours

    @property
    def prices(self) -> list:
        return self.model.prices_today

    @prices.setter
    def prices(self, val):
        self.model.prices_today = val
        if self.prices == self.prices_tomorrow:
            self.prices_tomorrow = []

    @property
    def prices_tomorrow(self) -> list:
        return self.model.prices_tomorrow

    @prices_tomorrow.setter
    def prices_tomorrow(self, val):
        ret = convert_none_list(val)
        if all([
            self.service.preserve_interim,
            ret is not None,
            ret != self.model.prices_tomorrow,
            len(ret) > 1
            ]):
            self.service.preserve_interim = False
        self.model.prices_tomorrow = ret

    @property
    def adjusted_average(self):
        return self.model.adjusted_average

    @adjusted_average.setter
    def adjusted_average(self, val):
        if val != self.model.adjusted_average:
            self.model.adjusted_average = val        
            self.update_prices(self.prices, self.prices_tomorrow)

    async def async_update_top_price(self, dyn_top_price=None) -> None:
        await self.model.options.async_set_absolute_top_price(dyn_top_price)
        self.update_prices(self.prices, self.prices_tomorrow)

    def update_prices(self, prices:dict = [], prices_tomorrow:dict=[]):
        self.prices = prices
        self.prices_tomorrow = prices_tomorrow
        self.service.update()

    async def async_get_average_kwh_price(self):
        ret = await self.async_get_charge_or_price()
        try:
            return round(sum(ret.values())/len(ret),2)
        except ZeroDivisionError as e:
                _LOGGER.warning(f"get_average_kwh_price_core could not be calculated: {e}")
        return 0
        
    async def async_get_total_charge(self, currentpeak:float) -> float:
        self.model.current_peak = currentpeak
        ret = await self.async_get_charge_or_price(True)
        return round(sum(ret.values()),1)

    async def async_get_charge_or_price(self, charge:bool = False) -> dict:
        hour = await self.service.async_set_hour()
        ret = {}

        async def async_looper_charge(h:int):
            if h in self.model.hours.dynamic_caution_hours:
                    ret[h] = self.model.hours.dynamic_caution_hours[h] * self.model.current_peak
            elif h in self.model.hours.non_hours:
                ret[h] = 0
            else:
                ret[h] = self.model.current_peak

        async def async_looper_price(h:int, tomorrow_active:bool):
            if h in self.model.hours.dynamic_caution_hours:
                    if tomorrow_active:
                        if h < hour and len(self.prices_tomorrow):
                            ret[h] = self.model.hours.dynamic_caution_hours[h] * self.prices_tomorrow[h]
                    if h >= hour:
                        ret[h] = self.model.hours.dynamic_caution_hours[h] * self.prices[h]
            elif h not in self.model.hours.non_hours:
                if h < hour and len(self.prices_tomorrow):
                    ret[h] = self.prices_tomorrow[h]
                if h >= hour:
                    ret[h] = self.prices[h]

        if self.prices_tomorrow is None or len(self.prices_tomorrow) < 1:
            for h in range(hour,24):
                if charge:
                    await async_looper_charge(h)
                else:
                    await async_looper_price(h, False)
        else:
            for h in range(hour,(hour+24)):
                h = h-24 if h > 23 else h
                if charge:
                    await async_looper_charge(h)
                else:
                    await async_looper_price(h, True)
        return ret

